#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <stdbool.h>
#include <ncurses.h>
#include <unistd.h>
#include <termios.h>
#include <limits.h>
#include <time.h>
#include <SDL2/SDL.h>
#include <SDL2/SDL_image.h>

#include "consts.h"
#include "base.h"
#include "display.h"
#include "edit.h"

void link_cells(cell* parent, cell* child) {
    if(parent != NULL) {
        if(parent->next_1 == NULL) {
            parent->next_1 = child;
        } else if(parent->next_2 == NULL) {
            parent->next_2 = child;
        } else if(parent->next_3 == NULL) {
            parent->next_3 = child;
        } else if(parent->next_4 == NULL) {
            parent->next_4 = child;
        } else {
            fprintf(stderr, "Unable to link cell : parent is at full capacity.\n");
        }
    };
    if(child != NULL) {
        if(child->prev_1 == NULL) {
            child->prev_1 = parent;
        } else if(child->prev_2 == NULL) {
            child->prev_2 = parent;
        } else if(child->prev_3 == NULL) {
            child->prev_3 = parent;
        } else if(child->prev_4 == NULL) {
            child->prev_4 = parent;
        } else {
            fprintf(stderr, "Unable to link cell : child is at full capacity.\n");
        }
    }
}

void unlink_cells(cell* parent, cell* child) {
    if(parent != NULL) {
        if(parent->next_1 == child) {
            parent->next_1 = NULL;
        } else if(parent->next_2 == child) {
            parent->next_2 = NULL;
        } else if(parent->next_3 == child) {
            parent->next_3 = NULL;
        } else if(parent->next_4 == child) {
            parent->next_4 = NULL;
        } else {
            fprintf(stderr, "Unable to unlink cell : no link found.\n");
        }
    };
    if(child != NULL) {
        if(child->prev_1 == parent) {
            child->prev_1 = NULL;
        } else if(child->prev_2 == parent) {
            child->prev_2 = NULL;
        } else if(child->prev_3 == parent) {
            child->prev_3 = NULL;
        } else if(child->prev_4 == parent) {
            child->prev_4 = NULL;
        } else {
            fprintf(stderr, "Unable to unlink cell : no link found.\n");
        }
    };
}

void append_cell(int cell_type, int X, int Y, cell* parent) {
    cell* new = malloc(sizeof(cell));
    new->id = parent->id;        // mostly for parkour
    new->coord_x = X;
    new->coord_y = Y;
    new->star_allowed = is_star_allowed(cell_type);
    new->type = (CELLTYPE)cell_type ;
    new->prev_1 = NULL ;
    new->prev_2 = NULL;
    new->prev_3 = NULL;
    new->prev_4 = NULL;
    new->next_1 = NULL;
    new->next_2 = NULL;
    new->next_3 = NULL;
    new->next_4 = NULL;
    char* meta = "none";

    link_cells(parent, new);
    cell_count[cell_type] += 1;
}

void append_cell_orthogonal(int cell_type, char dir, cell* parent) {
    if(dir == 'N' || dir == 'n') {
        append_cell(cell_type, parent->coord_x, parent->coord_y + tile_size * 2, parent);
    } else if(dir == 'E' || dir == 'e') {
        append_cell(cell_type, parent->coord_x + tile_size * 2, parent->coord_y, parent);
    } else if(dir == 'W' || dir == 'w') {
        append_cell(cell_type, parent->coord_x - tile_size * 2, parent->coord_y, parent);
    } else if(dir == 'S' || dir == 's') {
        append_cell(cell_type, parent->coord_x, parent->coord_y - tile_size * 2, parent);
    } else {
        fprintf(stderr, "ERROR : invalid direction (%d)", dir);
    }
}

void destroy_cell(cell* cl) {
    unlink_cells(cl, cl->next_1);
    unlink_cells(cl, cl->next_2);
    unlink_cells(cl, cl->next_3);
    unlink_cells(cl, cl->next_4);

    unlink_cells(cl->prev_1, cl);
    unlink_cells(cl->prev_2, cl);
    unlink_cells(cl->prev_3, cl);
    unlink_cells(cl->prev_4, cl);

    cell_count[(int)cl->type] -= 1;

    free(cl);
}

bool isAdding = false ;
int editBuffer = 0 ;
int phase = 0 ;

int editCellType = 0 ;
int editDirection = 0 ;

void playerEditActions(bool* finished) {
    SDL_Event event;
    while(SDL_PollEvent(&event)) {
        switch (event.type) {
            case SDL_QUIT:
                break;
            case SDL_KEYDOWN:
                switch (event.key.keysym.sym) {
                    case SDLK_1:
                        if(!isAdding && current->next_1 != NULL) {
                            current = current->next_1;
                        } else if (isAdding) {
                            editBuffer *= 10;
                            editBuffer += 1;
                        };
                        break;
                    case SDLK_2:
                        if(!isAdding && current->next_2 != NULL) {
                            current = current->next_2;
                        } else if (isAdding) {
                            editBuffer *= 10;
                            editBuffer += 2;
                        };
                        break;
                    case SDLK_3:
                        if(!isAdding && current->next_3 != NULL) {
                            current = current->next_3;
                        } else if (isAdding) {
                            editBuffer *= 10;
                            editBuffer += 3;
                        };
                        break;
                    case SDLK_4:
                        if(!isAdding && current->next_4 != NULL) {
                            current = current->next_4;
                        } else if (isAdding) {
                            editBuffer *= 10;
                            editBuffer += 4;
                        };
                        break;
                    case SDLK_0:
                        if(!isAdding && current->prev_1 != NULL) {
                            current = current->prev_1;
                        } else if (isAdding) {
                            editBuffer *= 10;
                        };
                        break;
                    case SDLK_9:
                        if(!isAdding && current->prev_2 != NULL) {
                            current = current->prev_2;
                        } else if (isAdding) {
                            editBuffer *= 10;
                            editBuffer += 9;
                        };
                        break;
                    case SDLK_8:
                        if(!isAdding && current->prev_3 != NULL) {
                            current = current->prev_3;
                        } else if (isAdding) {
                            editBuffer *= 10;
                            editBuffer += 8;
                        };
                        break;
                    case SDLK_7:
                        if(!isAdding && current->prev_4 != NULL) {
                            current = current->prev_4;
                        } else if (isAdding) {
                            editBuffer *= 10;
                            editBuffer += 7;
                        };
                        break;
                    case SDLK_6:
                        if (isAdding) {
                            editBuffer *= 10;
                            editBuffer += 6;
                        };
                        break;
                    case SDLK_5:
                        if (isAdding) {
                            editBuffer *= 10;
                            editBuffer += 5;
                        };
                        break;
                    case SDLK_MINUS:
                        if (isAdding) {
                            editBuffer = editBuffer / 10;
                        };
                        break;
                    case SDLK_SPACE:
                        if(isAdding) {
                            if(phase == 0) {
                                editCellType = editBuffer%n_cell_type;
                            } else if(phase == 1) {
                                editDirection = editBuffer%4;
                                phase = -1;
                                isAdding = false;
                                char dir ;
                                if(editDirection == 0) {
                                    dir = 's';
                                } else if(editDirection == 1) {
                                    dir = 'e';
                                } else if(editDirection == 2) {
                                    dir = 'n';
                                } else if(editDirection == 3) {
                                    dir = 'w';
                                };
                                append_cell_orthogonal(editCellType, dir, current);
                                printf("exit adding...\n");
                            };
                            phase += 1;
                            editBuffer = 0;
                        };
                        break;
                    case SDLK_p:
                        isAdding = true ;
                        phase = 0 ;
                        printf("now adding...\n");
                        break;
                    case SDLK_TAB:
                        *finished = true;
                        break;
                }
        }
    }
}

void draw_types(SDL_Renderer* renderer) {
    drawStringToRenderer(renderer, letters, "start",  __width__ - 6 * 75/2,  50, 75/2, 105/2);
    drawStringToRenderer(renderer, letters, "blue",   __width__ - 5 * 75/2, 100, 75/2, 105/2);
    drawStringToRenderer(renderer, letters, "red",    __width__ - 4 * 75/2, 150, 75/2, 105/2);
    drawStringToRenderer(renderer, letters, "event",  __width__ - 6 * 75/2, 200, 75/2, 105/2);
    drawStringToRenderer(renderer, letters, "vs",     __width__ - 3 * 75/2, 250, 75/2, 105/2);
    drawStringToRenderer(renderer, letters, "bowser", __width__ - 7 * 75/2, 300, 75/2, 105/2);
    drawStringToRenderer(renderer, letters, "chance", __width__ - 7 * 75/2, 350, 75/2, 105/2);
    drawStringToRenderer(renderer, letters, "boo",    __width__ - 4 * 75/2, 400, 75/2, 105/2);
    drawStringToRenderer(renderer, letters, "star",   __width__ - 5 * 75/2, 450, 75/2, 105/2);
    drawStringToRenderer(renderer, letters, "bank",   __width__ - 5 * 75/2, 500, 75/2, 105/2);
}

void draw_directions(SDL_Renderer* renderer) {
    drawStringToRenderer(renderer, letters, "north", __width__ - 6 * 75/2,  50, 75/2, 105/2);
    drawStringToRenderer(renderer, letters, "east",  __width__ - 5 * 75/2, 100, 75/2, 105/2);
    drawStringToRenderer(renderer, letters, "south", __width__ - 6 * 75/2, 150, 75/2, 105/2);
    drawStringToRenderer(renderer, letters, "west",  __width__ - 5 * 75/2, 200, 75/2, 105/2);
}

void edit_main(SDL_Renderer* renderer) {
    bool finished = false;
    while(!finished) {
        resetRenderer(renderer);
        draw_board(renderer, current->coord_x - tile_size * 7, current->coord_x + tile_size * 7, current->coord_y - tile_size * 7, current->coord_y + tile_size * 7, current);
        if(isAdding) {
            if(phase == 0) {
                drawStringToRenderer(renderer, letters, "choose type", 50, 50, 75/2, 105/2);
                draw_types(renderer);
            } else {
                drawStringToRenderer(renderer, letters, "choose direction", 50, 50, 75/2, 105/2);
                draw_directions(renderer);
            };
            drawNumberToRenderer(renderer, digits, editBuffer, 50, 150, 75/2, 105/2, 0);
        };
        updateRenderer(renderer);
        playerEditActions(&finished);
        usleep(20000);
    }
}